/* This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/** BusVoodoo high impedance (HiZ) default mode (code)
 *  @file busvoodoo_hiz.c
 *  @author King Kévin <kingkevin@cuvoodoo.info>
 *  @date 2018
 */
/* standard libraries */
#include <stdint.h> // standard integer types
#include <math.h> // math utilities

/* STM32 (including CM3) libraries */
#include <libopencm3/stm32/gpio.h> // general purpose input output library
#include <libopencm3/stm32/rcc.h> // real-time control clock library
#include <libopencm3/stm32/dbgmcu.h> // debug utilities
#include <libopencm3/stm32/desig.h> // design utilities
#include <libopencm3/stm32/dac.h> // DAC utilities

/* own libraries */
#include "global.h" // board definitions
#include "print.h" // printing utilities
#include "menu.h" // menu definitions
#include "busvoodoo_global.h" // BusVoodoo definitions
#include "busvoodoo_oled.h" // OLED utilities
#include "busvoodoo_hiz.h" // own definitions

#define BUSVOODOO_LV_DEFAULT (0.8*(1+30.0/10.0)) /**< default (when not driven) LV voltage regulator output voltage based on R1 and R2 */
#define BUSVOODOO_LV_TEST 2.5 /**< target LV output voltage to test if we can set control the LV voltage regulator */
#define BUSVOODOO_HV_DEFAULT (1.25*(1+100.0/10.0)) /**< default (when not driven) HV voltage regulator output voltage based on R1 and R2 */
#define BUSVOODOO_HV_TEST 12.0 /**< target HV output voltage to test if we can set control the HV voltage regulator */

/** setup HiZ mode
 *  @param[out] prefix terminal prompt prefix
 *  @param[in] line terminal prompt line to configure mode
 *  @return if setup is complete
 */
static bool busvoodoo_hiz_setup(char** prefix, const char* line)
{
	(void)line; // no configuration is required
	*prefix = "HiZ"; // set command line prefix
	led_blink(0, 1); // switch blue LED on to show we are ready
	busvoodoo_oled_text_left("HiZ"); // set mode title on OLED display
	const char* pinout_io[10] = {"GND", "3V3", "5V", "LV", NULL, NULL, NULL, NULL, NULL, NULL}; // HiZ mode pinout
	for (uint8_t i=0; i<LENGTH(pinout_io) && i<LENGTH(busvoodoo_global_pinout_io); i++) {
		busvoodoo_global_pinout_io[i] = pinout_io[i]; // set pin names
	}
	busvoodoo_oled_text_pinout(pinout_io, true); // set pinout on display
	busvoodoo_oled_update(); // update display to show text and pinout
	return true;	
}

/** exit HiZ mode
 */
static void busvoodoo_hiz_exit(void)
{
	// there is nothing to do
}

/** perform self tests
 *  @return if self tests passed
 */
static bool busvoodoo_hiz_test_self(void)
{
	bool to_return = false; // success of the self-test
	busvoodoo_safe_state(); // start from a safe state

	// get device information
	// get device identifier (DEV_ID)
	// 0x412: low-density, 16-32 kB flash
	// 0x410: medium-density, 64-128 kB flash
	// 0x414: high-density, 256-512 kB flash
	// 0x430: XL-density, 768-1024 kB flash
	// 0x418: connectivity
	if (0==(DBGMCU_IDCODE&DBGMCU_IDCODE_DEV_ID_MASK)) {
		printf("device identifier not set: this is probably a defective micro-controller\n");
	} else if (0x414!=(DBGMCU_IDCODE&DBGMCU_IDCODE_DEV_ID_MASK)) {
		printf("this (DEV_ID=%03x) is not a high-density device: a wrong micro-controller might have been used\n", (DBGMCU_IDCODE&DBGMCU_IDCODE_DEV_ID_MASK));
	}
	// ensure flash size is ok
	if (0xffff==DESIG_FLASH_SIZE) {
		printf("unknown flash size: this is probably a defective micro-controller\n");
	}

	// check 5V power rail
	float voltage = busvoodoo_vreg_get(BUSVOODOO_5V_CHANNEL); // get 5V power rail voltage
	if (voltage<4.0) {
		printf("5V power rail voltage is too low: %.2fV\n", voltage);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	} else if (voltage>5.5) {
		printf("5V power rail voltage is too high: %.2fV\n", voltage);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	}

	// check 3.3V power rail
	voltage = busvoodoo_vreg_get(BUSVOODOO_3V3_CHANNEL); // get 3.3V power rail voltage
	if (voltage<3.0) {
		printf("3.3V power rail voltage is too low: %.2fV\n", voltage);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	} else if (voltage>3.6) {
		printf("3.3V power rail voltage is too high: %.2fV\n", voltage);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	}

	// test 5V and 3.3V outputs
	busvoodoo_vout_switch(true); // enable Vout
	voltage = busvoodoo_vreg_get(BUSVOODOO_5V_CHANNEL); // get 5V power rail voltage
	if (voltage<4.0) {
		printf("5V power rail voltage is too low when 5V output is enabled: %.2fV\n", voltage);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	} else if (voltage>5.5) {
		printf("5V power rail voltage is too high when 5V output is enabled: %.2fV\n", voltage);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	}
	voltage = busvoodoo_vreg_get(BUSVOODOO_3V3_CHANNEL); // get 3.3V power rail voltage
	if (voltage<3.0) {
		printf("3.3V power rail voltage is too low when 3V3 output is enabled: %.2fV\n", voltage);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	} else if (voltage>3.6) {
		printf("3.3V power rail voltage is too high when 3V3 is enabled: %.2fV\n", voltage);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	}
	busvoodoo_vout_switch(false); // disable Vout
	
	// check LV voltage regulator
	voltage = busvoodoo_lv_set(0); // disable LV voltage regulator
	if (voltage>0.2) { // ensure the output is at 0V when the regulator is not enabled
		printf("LV voltage is %.2fV instead of 0V when the regulator is disabled\n", voltage);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	}
	gpio_set(GPIO(BUSVOODOO_LVEN_PORT), GPIO(BUSVOODOO_LVEN_PIN)); // enable LV voltage regulator
	sleep_ms(5); // let the voltage regulator start and voltage settle
	voltage = busvoodoo_vreg_get(BUSVOODOO_LV_CHANNEL); // get LV voltage
	// without being driven it should be around the default voltage
	if (voltage<BUSVOODOO_LV_DEFAULT-0.2) {
		printf("LV voltage is lower (%.2fV) than expected (%.2fV) when the regulator is enabled\n", voltage, BUSVOODOO_LV_DEFAULT);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	} else if (voltage>BUSVOODOO_LV_DEFAULT+0.2) {
		printf("LV voltage is higher (%.2fV) than expected (%.2fV) when the regulator is enabled\n", voltage, BUSVOODOO_LV_DEFAULT);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	}

	// check if we can control LV
	voltage = busvoodoo_lv_set(BUSVOODOO_LV_TEST); // get LV voltage
	if (voltage<BUSVOODOO_LV_TEST-0.2) {
		printf("LV voltage is lower (%.2fV) than set (%.2fV)\n", voltage, BUSVOODOO_LV_TEST);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	} else if (voltage>BUSVOODOO_LV_TEST+0.2) {
		printf("LV voltage is highed (%.2fV) than set (%.2fV)\n", voltage, BUSVOODOO_LV_TEST);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	}
	busvoodoo_lv_set(0); // disable LV voltage regulator

	// check HV voltage regulator
	if (busvoodoo_full) {
		voltage = busvoodoo_hv_set(0); // disable HV voltage regulator
		if (voltage>0.2) { // ensure the output is at 0V when the regulator is not enabled
			printf("HV voltage is %.2fV instead of 0V when the regulator is disabled\n", voltage);
#if DEBUG
			while (true);
#else
			goto error;
#endif
		}
		gpio_clear(GPIO(BUSVOODOO_HVEN_PORT), GPIO(BUSVOODOO_HVEN_PIN)); // enable HV voltage regulator
		sleep_ms(10); // let the voltage regulator start and voltage settle
		voltage = busvoodoo_vreg_get(BUSVOODOO_HV_CHANNEL); // get HV voltage
		// without being driven it should be around the default voltage
		if (voltage<BUSVOODOO_HV_DEFAULT-0.3) {
			printf("HV voltage is lower (%.2fV) than expected (%.2fV) when regulator is enabled\n", voltage, BUSVOODOO_HV_DEFAULT);
#if DEBUG
			while (true);
#else
			goto error;
#endif
		} else if (voltage>BUSVOODOO_HV_DEFAULT+0.3) {
			printf("HV voltage is higher (%.2fV) than expected (%.2fV) when regulator is enabled\n", voltage, BUSVOODOO_HV_DEFAULT);
#if DEBUG
			while (true);
#else
			goto error;
#endif
		}

		// check if we can control HV voltage regulator
		voltage = busvoodoo_hv_set(BUSVOODOO_HV_TEST); // set voltage on HV voltage regulator
		if (voltage<BUSVOODOO_HV_TEST-0.3) {
			printf("HV voltage is lower (%.2fV) than set (%.2fV)\n", voltage, BUSVOODOO_HV_TEST);
#if DEBUG
			while (true);
#else
			goto error;
#endif
		} else if (voltage>BUSVOODOO_HV_TEST+0.3) {
			printf("HV voltage is higher (%.2fV) than set (%.2fV)\n", voltage, BUSVOODOO_HV_TEST);
#if DEBUG
			while (true);
#else
			goto error;
#endif
		}
		voltage = busvoodoo_hv_set(0); // disable HV voltage regulator
	}
	
	// pull all pins down and ensure they are low
	for (uint8_t pin=0; pin<LENGTH(busvoodoo_io_ports) && pin<LENGTH(busvoodoo_io_pins); pin++) {
		gpio_set_mode(busvoodoo_io_ports[pin], GPIO_MODE_INPUT, GPIO_CNF_INPUT_PULL_UPDOWN, busvoodoo_io_pins[pin]); // set pin to input
		gpio_clear(busvoodoo_io_ports[pin], busvoodoo_io_pins[pin]); // pull down so it's not floating
	}
	for (uint8_t pin=0; pin<LENGTH(busvoodoo_io_ports) && pin<LENGTH(busvoodoo_io_pins) && pin<LENGTH(busvoodoo_io_names); pin++) {
		if (gpio_get(busvoodoo_io_ports[pin], busvoodoo_io_pins[pin])) { // ensure it really is low
			printf("signal %s is high although it is pulled low (internal)\n", busvoodoo_io_names[pin]); // warn user about the error
#if DEBUG
		while (true);
#else
		goto error;
#endif
		}
	}
	// pull all pins up and ensure they are high
	for (uint8_t pin=0; pin<LENGTH(busvoodoo_io_ports) && pin<LENGTH(busvoodoo_io_pins); pin++) {
		gpio_set_mode(busvoodoo_io_ports[pin], GPIO_MODE_INPUT, GPIO_CNF_INPUT_PULL_UPDOWN, busvoodoo_io_pins[pin]); // set pin to input
		gpio_set(busvoodoo_io_ports[pin], busvoodoo_io_pins[pin]); // pull up using internal pull-up
	}
	for (uint8_t pin=0; pin<LENGTH(busvoodoo_io_ports) && pin<LENGTH(busvoodoo_io_pins) && pin<LENGTH(busvoodoo_io_names); pin++) {
		if (!gpio_get(busvoodoo_io_ports[pin], busvoodoo_io_pins[pin])) { // ensure it really is high
			printf("signal %s is low although it is pulled up (internal)\n", busvoodoo_io_names[pin]); // warn user about the error
#if DEBUG
			while (true);
#else
			goto error;
#endif
		}
	}

	// set individual pin high and ensure only pins in the same group are at the same level
	for (uint8_t pin=0; pin<LENGTH(busvoodoo_io_ports) && pin<LENGTH(busvoodoo_io_pins); pin++) {
		gpio_set_mode(busvoodoo_io_ports[pin], GPIO_MODE_INPUT, GPIO_CNF_INPUT_PULL_UPDOWN, busvoodoo_io_pins[pin]); // set pin to input
		gpio_clear(busvoodoo_io_ports[pin], busvoodoo_io_pins[pin]); // pull down to ensure it is not high by accident
	}
	for (uint8_t pin1=0; pin1<LENGTH(busvoodoo_io_ports) && pin1<LENGTH(busvoodoo_io_pins) && pin1<LENGTH(busvoodoo_io_groups) && pin1<LENGTH(busvoodoo_io_names); pin1++) {
		gpio_set_mode(busvoodoo_io_ports[pin1], GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, busvoodoo_io_pins[pin1]); // set button pin to output
		gpio_set(busvoodoo_io_ports[pin1], busvoodoo_io_pins[pin1]); // set pin high
		for (uint8_t pin2=0; pin2<LENGTH(busvoodoo_io_ports) && pin2<LENGTH(busvoodoo_io_pins) && pin2<LENGTH(busvoodoo_io_groups) && pin2<LENGTH(busvoodoo_io_names); pin2++) {
			if (busvoodoo_io_groups[pin1]==busvoodoo_io_groups[pin2] && !gpio_get(busvoodoo_io_ports[pin2], busvoodoo_io_pins[pin2])) {
				printf("signal %s of I/O-%u is low while it should be set high by signal %s of I/O-%u\n", busvoodoo_io_names[pin2], busvoodoo_io_groups[pin2], busvoodoo_io_names[pin1], busvoodoo_io_groups[pin1]); // warn user about the error
#if DEBUG
				while (true);
#else
				goto error;
#endif
			} else if (busvoodoo_io_groups[pin1]!=busvoodoo_io_groups[pin2] && gpio_get(busvoodoo_io_ports[pin2], busvoodoo_io_pins[pin2])) {
				printf("signal %s of I/O-%u is high while it should not be set high by signal %s of I/O-%u\n", busvoodoo_io_names[pin2], busvoodoo_io_groups[pin2], busvoodoo_io_names[pin1], busvoodoo_io_groups[pin1]); // warn user about the error
#if DEBUG
				while (true);
#else
				goto error;
#endif
			}
		}
		gpio_set_mode(busvoodoo_io_ports[pin1], GPIO_MODE_INPUT, GPIO_CNF_INPUT_PULL_UPDOWN, busvoodoo_io_pins[pin1]); // set pin back to input
		gpio_clear(busvoodoo_io_ports[pin1], busvoodoo_io_pins[pin1]); // pull pin back down
	}
	// set individual pin low and ensure only pins in the same group are at the same level
	for (uint8_t pin=0; pin<LENGTH(busvoodoo_io_ports) && pin<LENGTH(busvoodoo_io_pins); pin++) {
		gpio_set_mode(busvoodoo_io_ports[pin], GPIO_MODE_INPUT, GPIO_CNF_INPUT_PULL_UPDOWN, busvoodoo_io_pins[pin]); // set pin to input
		gpio_set(busvoodoo_io_ports[pin], busvoodoo_io_pins[pin]); // pull up to ensure it is not low by accident
	}
	for (uint8_t pin1=0; pin1<LENGTH(busvoodoo_io_ports) && pin1<LENGTH(busvoodoo_io_pins) && pin1<LENGTH(busvoodoo_io_groups) && pin1<LENGTH(busvoodoo_io_names); pin1++) {
		gpio_set_mode(busvoodoo_io_ports[pin1], GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, busvoodoo_io_pins[pin1]); // set button pin to output
		gpio_clear(busvoodoo_io_ports[pin1], busvoodoo_io_pins[pin1]); // set pin low
		for (uint8_t pin2=0; pin2<LENGTH(busvoodoo_io_ports) && pin2<LENGTH(busvoodoo_io_pins) && pin2<LENGTH(busvoodoo_io_groups) && pin2<LENGTH(busvoodoo_io_names); pin2++) {
			if (busvoodoo_io_groups[pin1]==busvoodoo_io_groups[pin2] && gpio_get(busvoodoo_io_ports[pin2], busvoodoo_io_pins[pin2])) {
				printf("signal %s of I/O-%u is high while it should be set low by signal %s of I/O-%u\n", busvoodoo_io_names[pin2], busvoodoo_io_groups[pin2], busvoodoo_io_names[pin1], busvoodoo_io_groups[pin1]); // warn user about the error
#if DEBUG
				while (true);
#else
				goto error;
#endif
			} else if (busvoodoo_io_groups[pin1]!=busvoodoo_io_groups[pin2] && !gpio_get(busvoodoo_io_ports[pin2], busvoodoo_io_pins[pin2])) {
				printf("signal %s of I/O-%u is low while it should not be set low by signal %s of I/O-%u\n", busvoodoo_io_names[pin2], busvoodoo_io_groups[pin2], busvoodoo_io_names[pin1], busvoodoo_io_groups[pin1]); // warn user about the error
#if DEBUG
				while (true);
#else
				goto error;
#endif
			}
		}
		gpio_set_mode(busvoodoo_io_ports[pin1], GPIO_MODE_INPUT, GPIO_CNF_INPUT_PULL_UPDOWN, busvoodoo_io_pins[pin1]); // set pin back to input
		gpio_set(busvoodoo_io_ports[pin1], busvoodoo_io_pins[pin1]); // pull pin back up
	}

	// test 5V pull-up
	for (uint8_t pin=0; pin<LENGTH(busvoodoo_io_ports) && pin<LENGTH(busvoodoo_io_pins); pin++) {
		gpio_set_mode(busvoodoo_io_ports[pin], GPIO_MODE_INPUT, GPIO_CNF_INPUT_PULL_UPDOWN, busvoodoo_io_pins[pin]); // set pin to input
		gpio_clear(busvoodoo_io_ports[pin], busvoodoo_io_pins[pin]); // pull down to ensure it is not high by accident
	}
	busvoodoo_lv_set(5.0); // set LV voltage regulator to use the 5.0V power rail
	voltage = busvoodoo_embedded_pullup(true); // enable 5V pull-up
	if (voltage<4.0) {
		printf("5V power rail voltage is too low when used to pull up: %.2fV\n", voltage);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	} else if (voltage>5.5) {
		printf("5V power rail voltage is too high when used to pull up: %.2fV\n", voltage);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	}
	for (uint8_t pin=0; pin<LENGTH(busvoodoo_io_ports) && pin<LENGTH(busvoodoo_io_pins) && pin<LENGTH(busvoodoo_io_names); pin++) {
		if (!gpio_get(busvoodoo_io_ports[pin], busvoodoo_io_pins[pin])) { // ensure it really is high
			printf("signal %s is low although it is pulled up by 5V (embedded)\n", busvoodoo_io_names[pin]); // warn user about the error
#if DEBUG
			while (true);
#else
			goto error;
#endif
		}
	}
	busvoodoo_embedded_pullup(false); // disable pull-ups
	voltage = busvoodoo_lv_set(0); // disable LV voltage regulator

	// test LV pull-up set to 3.3V
	for (uint8_t pin=0; pin<LENGTH(busvoodoo_io_ports) && pin<LENGTH(busvoodoo_io_pins); pin++) {
		gpio_set_mode(busvoodoo_io_ports[pin], GPIO_MODE_INPUT, GPIO_CNF_INPUT_PULL_UPDOWN, busvoodoo_io_pins[pin]); // set pin to input
		gpio_clear(busvoodoo_io_ports[pin], busvoodoo_io_pins[pin]); // pull down to ensure it is not high by accident
	}
	busvoodoo_lv_set(3.3); // set LV voltage regulator
	voltage = busvoodoo_embedded_pullup(3.3); // enable pull-up with adjustable regulator
	if (voltage<BUSVOODOO_LV_DEFAULT-0.2) {
		printf("LV voltage is lower (%.2fV) than expected (%.2fV) when used to pull up\n", voltage, BUSVOODOO_LV_DEFAULT);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	} else if (voltage>BUSVOODOO_LV_DEFAULT+0.2) {
		printf("LV voltage is higher (%.2fV) than expected (%.2fV) when used to pull up\n", voltage, BUSVOODOO_LV_DEFAULT);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	}
	for (uint8_t pin=0; pin<LENGTH(busvoodoo_io_ports) && pin<LENGTH(busvoodoo_io_pins) && pin<LENGTH(busvoodoo_io_names); pin++) {
		if (!gpio_get(busvoodoo_io_ports[pin], busvoodoo_io_pins[pin])) { // ensure it really is high
			printf("signal %s is low although it is pulled up by LV (embedded)\n", busvoodoo_io_names[pin]); // warn user about the error
#if DEBUG
			while (true);
#else
			goto error;
#endif
		}
	}
	busvoodoo_lv_set(0); // disable LV voltage regulator
	busvoodoo_embedded_pullup(false); // disable pull-ups

	to_return = true; // all tests are successful
#if DEBUG
#else
error:
#endif
	busvoodoo_safe_state(); // set back to safe state
	if (!to_return) {
		printf("the test procedure has been aborted for safety reasons\n");
	}
	
	return to_return;
}

/** test if signals are soldered correctly to the connector pins 
 *  @return if pin test passed
 */
static bool busvoodoo_hiz_test_pins(void)
{
	bool to_return = false; // test result to return
	busvoodoo_safe_state(); // start from safe state with all outputs switched off

	const char* lv_to = "connect I/O pin 4 to "; // most outputs will be tested using LV ADC
	const char* lv_high = "pin 4 is high while it should be low, other pin might be defective\n"; // error message if expect low output is high

	// test GND on pin 1 by shorting LV_CTL to ground (through 40k resistor) and measure short
	gpio_set(GPIO(BUSVOODOO_LVCTL_PORT), GPIO(BUSVOODOO_LVCTL_PIN)); // set pin high
	gpio_set_mode(GPIO(BUSVOODOO_LVCTL_PORT), GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, GPIO(BUSVOODOO_LVCTL_PIN)); // set LV control pin as output
	printf("%sI/O pin 1\n", lv_to);
	do {
		sleep_ms(100); // wait for user to make connection
	} while (busvoodoo_vreg_get(BUSVOODOO_LV_CHANNEL)>0.2 && !user_input_available); // wait until pin is shorted to ground
	if (user_input_available) { // user interruption
		goto error;
	}
	gpio_clear(GPIO(BUSVOODOO_LVCTL_PORT), GPIO(BUSVOODOO_LVCTL_PIN)); // set pin low
	gpio_set_mode(GPIO(BUSVOODOO_LVCTL_PORT), GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, GPIO(BUSVOODOO_LVCTL_PIN)); // set LV control pin as output
	led_toggle(); // notify user test is almost almost
	sleep_ms(100); // wait for voltage to settle an debounce
	if (busvoodoo_vreg_get(BUSVOODOO_LV_CHANNEL)>0.2) {
		printf(lv_high);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	}
	gpio_set_mode(GPIO(BUSVOODOO_LVCTL_PORT), GPIO_MODE_INPUT, GPIO_CNF_INPUT_ANALOG, GPIO(BUSVOODOO_LVCTL_PIN)); // set LV control pin back to analog input for DAC
	led_toggle(); // notify user test is complete
	
	// test 5V output on pin 2
	gpio_clear(GPIO(BUSVOODOO_VOUTEN_PORT), GPIO(BUSVOODOO_VOUTEN_PIN)); // enable Vout
	printf("%sI/O pin 2\n", lv_to);
	do {
		sleep_ms(100); // wait for user to make connection
	} while (busvoodoo_vreg_get(BUSVOODOO_LV_CHANNEL)<0.2 && !user_input_available); // wait until pin is connected
	if (user_input_available) { // user interruption
		goto error;
	}
	gpio_set(GPIO(BUSVOODOO_VOUTEN_PORT), GPIO(BUSVOODOO_VOUTEN_PIN)); // disable Vout
	led_toggle(); // notify user test is almost complete
	sleep_ms(100); // wait for voltage to settle and debounce
	if (busvoodoo_vreg_get(BUSVOODOO_LV_CHANNEL)>0.2) {
		printf(lv_high);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	}
	led_toggle(); // notify user test is complete

	// test 3.3V output on pin 3
	gpio_clear(GPIO(BUSVOODOO_VOUTEN_PORT), GPIO(BUSVOODOO_VOUTEN_PIN)); // enable Vout
	printf("%sI/O pin 3\n", lv_to);
	do {
		sleep_ms(100); // wait for user to make connection
	} while ((busvoodoo_vreg_get(BUSVOODOO_LV_CHANNEL)<0.2 || busvoodoo_vreg_get(BUSVOODOO_LV_CHANNEL)>3.5) && !user_input_available); // wait until pin is connected
	if (user_input_available) { // user interruption
		goto error;
	}
	gpio_set(GPIO(BUSVOODOO_VOUTEN_PORT), GPIO(BUSVOODOO_VOUTEN_PIN)); // disable Vout
	led_toggle(); // notify user test is almost complete
	sleep_ms(100); // wait for voltage to settle and debounce
	if (busvoodoo_vreg_get(BUSVOODOO_LV_CHANNEL)>0.2) {
		printf(lv_high);
#if DEBUG
		while (true);
#else
		goto error;
#endif
	}
	led_toggle(); // notify user test is complete

	// test I/O pins
	for (uint8_t io=1; io<=6; io++) { // test each I/O pin
		for (uint8_t pin=0; pin<LENGTH(busvoodoo_io_ports) && pin<LENGTH(busvoodoo_io_pins) && pin<LENGTH(busvoodoo_io_groups); pin++) { // look for a pin mapped on this I/O
			if (busvoodoo_io_groups[pin]==io) {
				gpio_set(busvoodoo_io_ports[pin], busvoodoo_io_pins[pin]); // set pin high
				gpio_set_mode(busvoodoo_io_ports[pin], GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, busvoodoo_io_pins[pin]); // set pin to output
				printf("%sI/O pin %u\n", lv_to, io+4);
				do {
					sleep_ms(100); // wait for user to make connection
				} while (busvoodoo_vreg_get(BUSVOODOO_LV_CHANNEL)<0.2 && !user_input_available); // wait until pin is connected
				if (user_input_available) { // user interruption
					goto error;
				}
				gpio_clear(busvoodoo_io_ports[pin], busvoodoo_io_pins[pin]); // set pin low
				led_toggle(); // notify user test is almost complete
				sleep_ms(100); // wait for voltage to settle and debounce
				if (busvoodoo_vreg_get(BUSVOODOO_LV_CHANNEL)>0.2) {
					printf(lv_high);
#if DEBUG
					while (true);
#else
					goto error;
#endif
				}
				gpio_set_mode(busvoodoo_io_ports[pin], GPIO_MODE_INPUT, GPIO_CNF_INPUT_FLOAT, busvoodoo_io_pins[pin]); // set pin back to input
				led_toggle(); // notify user test is complete
				break; // stop looking for pin
			}
		}
	}

	if (busvoodoo_full) {
		// test HV output on RS/CAN pin 1
		double voltage = busvoodoo_vreg_get(BUSVOODOO_3V3_CHANNEL); // get reference voltage
		uint16_t dac_set = BUSVOODOO_HV_SET(5.0)/voltage*4095; // DAC value corresponding to the voltage
		dac_load_data_buffer_single(dac_set, RIGHT12, BUSVOODOO_HVCTL_CHANNEL); // set output so the voltage regulator is set to desired output voltage
		dac_software_trigger(BUSVOODOO_HVCTL_CHANNEL); // transfer the value to the DAC
		dac_enable(BUSVOODOO_HVCTL_CHANNEL); // enable DAC
		gpio_clear(GPIO(BUSVOODOO_HVEN_PORT), GPIO(BUSVOODOO_HVEN_PIN)); // enable HV voltage regulator
		printf("%sRS/CAN pin 1\n", lv_to);
		do {
			sleep_ms(100); // wait for user to make connection
		} while (busvoodoo_vreg_get(BUSVOODOO_LV_CHANNEL)<0.2 && !user_input_available); // wait until pin is connected
		if (user_input_available) { // user interruption
			goto error;
		}
		gpio_set(GPIO(BUSVOODOO_HVEN_PORT), GPIO(BUSVOODOO_HVEN_PIN)); // disable HV voltage regulator
		dac_disable(BUSVOODOO_HVCTL_CHANNEL); // disable HV control
		led_toggle(); // notify user test is almost complete
		sleep_ms(100); // wait for voltage to settle (and debounce)
		if (busvoodoo_vreg_get(BUSVOODOO_LV_CHANNEL)>0.2) {
			printf(lv_high);
#if DEBUG
			while (true);
#else
			goto error;
#endif
		}
		led_toggle(); // notify user test is complete

		// test RS-232 port (with itself)
		rcc_periph_clock_enable(RCC_GPIO(BUSVOODOO_RS232_EN_PORT)); // enable clock for GPIO domain
		gpio_clear(GPIO(BUSVOODOO_RS232_EN_PORT), GPIO(BUSVOODOO_RS232_EN_PIN)); // set low to enable receiver
		gpio_set_mode(GPIO(BUSVOODOO_RS232_EN_PORT), GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_OPENDRAIN, GPIO(BUSVOODOO_RS232_EN_PIN)); // set pin as output (open-drain pulled high to disable receiver)
		rcc_periph_clock_enable(RCC_GPIO(BUSVOODOO_RS232_SHDN_PORT)); // enable clock for GPIO domain
		gpio_set(GPIO(BUSVOODOO_RS232_SHDN_PORT), GPIO(BUSVOODOO_RS232_SHDN_PIN)); // set high to enable transmitter
		gpio_set_mode(GPIO(BUSVOODOO_RS232_SHDN_PORT), GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, GPIO(BUSVOODOO_RS232_SHDN_PIN)); // set pin as output (push-pull pulled low to disable transmitter)
		rcc_periph_clock_enable(RCC_GPIO(BUSVOODOO_RS232_TX_PORT)); // enable clock for GPIO
	    gpio_set_mode(GPIO(BUSVOODOO_RS232_TX_PORT), GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, GPIO(BUSVOODOO_RS232_TX_PIN)); // set pin as output (push-pull)
		rcc_periph_clock_enable(RCC_GPIO(BUSVOODOO_RS232_RX_PORT)); // enable clock for GPIO
		gpio_set_mode(GPIO(BUSVOODOO_RS232_RX_PORT), GPIO_MODE_INPUT, GPIO_CNF_INPUT_PULL_UPDOWN, GPIO(BUSVOODOO_RS232_RX_PIN)); // set pin as input (with pull resistors)
		// start by setting low since unconnected (pulled to ground by 3 kO) is considered as high
		gpio_clear(GPIO(BUSVOODOO_RS232_TX_PORT), GPIO(BUSVOODOO_RS232_TX_PIN)); // set low
		gpio_set(GPIO(BUSVOODOO_RS232_RX_PORT), GPIO(BUSVOODOO_RS232_RX_PIN)); // pull high to avoid false negative
		sleep_ms(5);
		printf("connect RS/CAN pin 2 to RS/CAN pin 3\n");
		do {
			sleep_ms(100); // wait for user to make connection
		} while (gpio_get(GPIO(BUSVOODOO_RS232_RX_PORT), GPIO(BUSVOODOO_RS232_RX_PIN)) && !user_input_available); // wait until pin is connected
		if (user_input_available) { // user interruption
			goto error;
		}
		gpio_set(GPIO(BUSVOODOO_RS232_TX_PORT), GPIO(BUSVOODOO_RS232_TX_PIN)); // set high
		gpio_clear(GPIO(BUSVOODOO_RS232_RX_PORT), GPIO(BUSVOODOO_RS232_RX_PIN)); // pull low to avoid false negative
		led_toggle(); // notify user test is almost complete
		sleep_ms(100); // wait for voltage to settle and debounce
		if (!gpio_get(GPIO(BUSVOODOO_RS232_RX_PORT), GPIO(BUSVOODOO_RS232_RX_PIN))) { // check if RX is set low by TX
			printf("CAN/RS pin 2 is high while it should be set low by pin 3\n");
#if DEBUG
			while (true);
#else
			goto error;
#endif
		}
		gpio_set_mode(GPIO(BUSVOODOO_RS232_TX_PORT), GPIO_MODE_INPUT, GPIO_CNF_INPUT_FLOAT, GPIO(BUSVOODOO_RS232_TX_PIN)); // free pin
		gpio_set_mode(GPIO(BUSVOODOO_RS232_RX_PORT), GPIO_MODE_INPUT, GPIO_CNF_INPUT_FLOAT, GPIO(BUSVOODOO_RS232_RX_PIN)); // free pin
		led_toggle(); // notify user test is complete

		rcc_periph_clock_enable(RCC_GPIO(BUSVOODOO_RS232_RTS_PORT)); // enable clock for GPIO
	    gpio_set_mode(GPIO(BUSVOODOO_RS232_RTS_PORT), GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, GPIO(BUSVOODOO_RS232_RTS_PIN)); // set pin as output (push-pull)
		rcc_periph_clock_enable(RCC_GPIO(BUSVOODOO_RS232_CTS_PORT)); // enable clock for GPIO
		gpio_set_mode(GPIO(BUSVOODOO_RS232_CTS_PORT), GPIO_MODE_INPUT, GPIO_CNF_INPUT_PULL_UPDOWN, GPIO(BUSVOODOO_RS232_CTS_PIN)); // set pin as input (with pull resistors)
		// start by setting low since unconnected (pulled to ground by 3 kO) is considered as high
		gpio_clear(GPIO(BUSVOODOO_RS232_RTS_PORT), GPIO(BUSVOODOO_RS232_RTS_PIN)); // set low
		gpio_set(GPIO(BUSVOODOO_RS232_CTS_PORT), GPIO(BUSVOODOO_RS232_CTS_PIN)); // pull high to avoid false negative
		printf("connect RS/CAN pin 4 to RS/CAN pin 5\n");
		do {
			sleep_ms(100); // wait for user to make connection
		} while (gpio_get(GPIO(BUSVOODOO_RS232_CTS_PORT), GPIO(BUSVOODOO_RS232_CTS_PIN)) && !user_input_available); // wait until pin is connected
		if (user_input_available) { // user interruption
			goto error;
		}
		gpio_set(GPIO(BUSVOODOO_RS232_RTS_PORT), GPIO(BUSVOODOO_RS232_RTS_PIN)); // set high
		gpio_clear(GPIO(BUSVOODOO_RS232_CTS_PORT), GPIO(BUSVOODOO_RS232_CTS_PIN)); // pull low to avoid false negative
		led_toggle(); // notify user test is almost complete
		sleep_ms(100); // wait for voltage to settle an debounce
		if (!gpio_get(GPIO(BUSVOODOO_RS232_CTS_PORT), GPIO(BUSVOODOO_RS232_CTS_PIN))) { // check if CTS is set high by RTS
			printf("CAN/RS pin 5 is high while it should be set low by pin 4\n");
#if DEBUG
			while (true);
#else
			goto error;
#endif
		}
		gpio_set_mode(GPIO(BUSVOODOO_RS232_RTS_PORT), GPIO_MODE_INPUT, GPIO_CNF_INPUT_FLOAT, GPIO(BUSVOODOO_RS232_RTS_PIN)); // free pin
		gpio_set_mode(GPIO(BUSVOODOO_RS232_CTS_PORT), GPIO_MODE_INPUT, GPIO_CNF_INPUT_FLOAT, GPIO(BUSVOODOO_RS232_CTS_PIN)); // free pin
		led_toggle(); // notify user test is complete

		gpio_set(GPIO(BUSVOODOO_RS232_EN_PORT), GPIO(BUSVOODOO_RS232_EN_PIN)); // set high to disable receiver
		gpio_clear(GPIO(BUSVOODOO_RS232_SHDN_PORT), GPIO(BUSVOODOO_RS232_SHDN_PIN)); // set low to disable transmitter
	}
	
	to_return = true; // all tests passed
error:
	busvoodoo_safe_state(); // go back to safe state
	if (user_input_available) { // user interrupted the test
		printf("user interrupted the test\n");
		to_return = true; // we don't consider this as error
	}
	if (!to_return) { // test has not been completed
		printf("the test procedure has been aborted for safety reasons\n");
	}

	return to_return;
}

// command handlers
/** command to perform board self-test
 *  @param[in] argument no argument required
 */
static void busvoodoo_hiz_command_test_self(void* argument)
{
	(void)argument; // we won't use the argument
	printf("performing self-test\n");
	printf("remove all cables from connectors and press space to start (or any other key to abort)\n");
	if (' '==user_input_get()) { // space entered
		if (busvoodoo_hiz_test_self()) { // perform self-test
			led_blink(0, 1.0); // show blue OK LED
			printf("self-test succeeded\n"); // notify user
		} else {
			led_blink(0.5, 0.5); // show error on LEDs
			printf("self-test failed\n"); // notify user
		}
	} else {
		printf("self-test aborted\n");
	}
}

/** command to perform pins test
 *  @param[in] argument no argument required
 */
static void busvoodoo_hiz_command_test_pins(void* argument)
{
	(void)argument; // we won't use the argument
	printf("performing pins test\n");
	printf("remove all cables from connectors and press space to start (or any other key to abort)\n");
	if (' '==user_input_get()) { // space entered
		if (busvoodoo_hiz_test_pins()) { // perform pin test
			led_blink(0, 1.0); // show blue OK LED
			printf("pins test succeeded\n"); // notify user
		} else {
			led_blink(0.5, 0.5); // show error on LEDs
			printf("pins test failed\n"); // notify user
		}
	} else {
		printf("pins test aborted\n");
	}
}

static const struct menu_command_t busvoodoo_hiz_commands[] = {
	{
		's',
		"self-test",
		"perform board self-test",
		MENU_ARGUMENT_NONE,
		NULL,
		&busvoodoo_hiz_command_test_self,
	},
	{
		't',
		"pins-test",
		"perform connector pins test",
		MENU_ARGUMENT_NONE,
		NULL,
		&busvoodoo_hiz_command_test_pins,
	},
};

struct busvoodoo_mode_t busvoodoo_hiz_mode = {
	"hiz",
	"High Impedance (Z)",
	&busvoodoo_hiz_setup,
	busvoodoo_hiz_commands,
	LENGTH(busvoodoo_hiz_commands),
	&busvoodoo_hiz_exit,
};
